local Ellyb = Ellyb(...);

local configureHoverFrame = Questo_API.lib.configureHoverFrame;
local setTooltipForSameFrame, setTooltipAll = Questo_API.lib.setTooltipForSameFrame, Questo_API.lib.setTooltipAll;
local refreshTooltipForFrame = Questo_RefreshTooltipForFrame;
local Questo_MainTooltip = Questo_MainTooltip;
local log = Questo_API.lib.log;
local getTextureString, colorCodeFloat = Questo_API.lib.getTextureString, Questo_API.lib.colorCodeFloat;
local getId = Questo_API.lib.generateID;
local loc = Questo_API.locale.getText;
local format = format;
local getQuestIcon, getQuestActiveIcon = Questo_API.getQuestIcon, Questo_API.getQuestActiveIcon;
local selectFirstGossip, 	selectMultipleGossip = Questo_API.selectFirstGossip, Questo_API.selectMultipleGossip;
local selectMultipleRewards, selectFirstGreetingActive = Questo_API.selectMultipleRewards, Questo_API.selectFirstGreetingActive;
local hideQuestoFrame = Questo_API.layout.hideQuestoFrame;
local hideQuestRewardFrameIfNeed = Questo_API.layout.hideQuestRewardFrameIfNeed;
local debug = Questo_API.debug;
local tsize = Questo_API.lib.tsize;

local faction, faction_loc = UnitFactionGroup("player");
local pairs, CreateFrame, wipe, type, tinsert, after, select, huge = pairs, CreateFrame, wipe, type, tinsert, C_Timer.After, select, math.huge;
local ChatTypeInfo = ChatTypeInfo;
local UnitIsUnit, UnitExists, DeclineQuest, AcceptQuest, AcknowledgeAutoAcceptQuest = UnitIsUnit, UnitExists, DeclineQuest, AcceptQuest, AcknowledgeAutoAcceptQuest;
local IsQuestCompletable, CompleteQuest, CloseQuest, GetQuestLogTitle = IsQuestCompletable, CompleteQuest, CloseQuest, GetQuestLogTitle;
local GetNumQuestChoices, GetQuestReward, GetQuestLogSelection = GetNumQuestChoices, GetQuestReward, GetQuestLogSelection;
local GetQuestLogQuestText, GetGossipAvailableQuests, GetGossipActiveQuests = GetQuestLogQuestText, GetGossipAvailableQuests, GetGossipActiveQuests;
local GetNumGossipOptions, GetNumGossipAvailableQuests, GetNumGossipActiveQuests = GetNumGossipOptions, GetNumGossipAvailableQuests, GetNumGossipActiveQuests;
local GetQuestItemInfo, GetNumQuestItems, GetGossipOptions = GetQuestItemInfo, GetNumQuestItems, GetGossipOptions;
local GetObjectiveText, GetCoinTextureString, GetRewardXP = GetObjectiveText, GetCoinTextureString, GetRewardXP;
local GetQuestItemLink, GetNumQuestRewards, GetRewardMoney, GetNumRewardCurrencies = GetQuestItemLink, GetNumQuestRewards, GetRewardMoney, GetNumRewardCurrencies;
local GetRewardSkillPoints = GetRewardSkillPoints;
local GetAvailableQuestInfo, GetNumAvailableQuests, GetNumActiveQuests = GetAvailableQuestInfo, GetNumAvailableQuests, GetNumActiveQuests;
local GetAvailableTitle, GetActiveTitle, CloseGossip = GetAvailableTitle, GetActiveTitle, CloseGossip;
local GetProgressText, GetTitleText, GetGreetingText = GetProgressText, GetTitleText, GetGreetingText;
local GetGossipText, GetRewardText, GetQuestText = GetGossipText, GetRewardText, GetQuestText;
local GetItemInfo, GetContainerNumSlots, GetContainerItemLink, EquipItemByName = GetItemInfo, GetContainerNumSlots, GetContainerItemLink, EquipItemByName;
local InCombatLockdown, GetInventorySlotInfo, GetInventoryItemLink = InCombatLockdown, GetInventorySlotInfo, GetInventoryItemLink;
local GetQuestCurrencyInfo, GetMaxRewardCurrencies, GetRewardTitle = GetQuestCurrencyInfo, GetMaxRewardCurrencies, GetRewardTitle;
local GetFollowerInfo = C_Garrison.GetFollowerInfo;
local GetQuestMoneyToGet, GetMoney, GetNumQuestCurrencies = GetQuestMoneyToGet, GetMoney, GetNumQuestCurrencies;
local GetSuggestedGroupNum = GetSuggestedGroupNum;
local UnitIsDead = UnitIsDead;
local QuestIsFromAreaTrigger, QuestGetAutoAccept = QuestIsFromAreaTrigger, QuestGetAutoAccept;
local BreakUpLargeNumbers = BreakUpLargeNumbers;

local Questo_NPCFrameObjectives, Questo_NPCFrameObjectivesNo, Questo_NPCFrameObjectivesYes = Questo_NPCFrameObjectives, Questo_NPCFrameObjectivesNo, Questo_NPCFrameObjectivesYes;
local Questo_NPCFrameObjectivesImage = Questo_NPCFrameObjectivesImage;
local Questo_NPCFrameRewardsItemIcon, Questo_NPCFrameRewardsItem, Questo_NPCFrameRewards = Questo_NPCFrameRewardsItemIcon, Questo_NPCFrameRewardsItem, Questo_NPCFrameRewards;
local Questo_NPCFrame, Questo_NPCFrameChatNextText = Questo_NPCFrame, Questo_NPCFrameChatNextText;
local Questo_NPCFrameChat, Questo_NPCFrameChatText = Questo_NPCFrameChat, Questo_NPCFrameChatText;
local Questo_NPCFrameChatNext, Questo_NPCFrameChatPrevious = Questo_NPCFrameChatNext, Questo_NPCFrameChatPrevious;
local Questo_NPCFrameConfigButton, Questo_NPCFrameObjectivesContent = Questo_NPCFrameConfigButton, Questo_NPCFrameObjectivesContent;
local Questo_NPCFrameGossipChoices = Questo_NPCFrameGossipChoices;
local Dialogs = Questo_API.dialogs;
local DialogsButtons = Questo_API.dialogs.buttons;
local DialogsScrollFrame = Questo_API.dialogs.scrollFrame;
local Rewards = Questo_API.rewards;
local RewardsButtons = Questo_API.rewards.buttons;

local OPTIONS_MARGIN, OPTIONS_TOP = 175, -175;
local GOSSIP_DELAY = 0.2;
local gossipColor = "|cffffffff";
local EVENT_INFO;
local eventHandlers = {};
local BONUS_SKILLPOINTS = BONUS_SKILLPOINTS;
local ITEM_QUALITY_COLORS = ITEM_QUALITY_COLORS;
local REQUIRED_MONEY = REQUIRED_MONEY;
local QUEST_SUGGESTED_GROUP_NUM, QUEST_OBJECTIVES = QUEST_SUGGESTED_GROUP_NUM, QUEST_OBJECTIVES;

local SEAL_QUESTS = {
	[40519] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_VARIAN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal"},
	[43926] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_VOLJIN.."|r", sealAtlas = "Quest-Horde-WaxSeal"},
	[47221] = { bgAtlas = "QuestBG-TheHandofFate", },
	[47835] = { bgAtlas = "QuestBG-TheHandofFate", },
	[49929] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal" },
	[49930] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_SYLVANAS_WINDRUNNER.."|r", sealAtlas = "Quest-Horde-WaxSeal" },
	[50476] = { bgAtlas = "QuestBG-Horde", sealAtlas = "Quest-Horde-WaxSeal" },
	[46727] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal" },
	[50668] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_SYLVANAS_WINDRUNNER.."|r", sealAtlas = "Quest-Horde-WaxSeal"},

	[51795] = { bgAtlas = "QuestBG-Alliance" },
	[52058] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal"},

	[51796] = { bgAtlas = "QuestBG-Horde" },

	[53372] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_SYLVANAS_WINDRUNNER.."|r", sealAtlas = "Quest-Horde-WaxSeal"},
	[53370] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal"},
};
local EXCEPTION_QUESTS = {
	[53029] = true,
	[53026] = true,
	[51211] = true,
	[52428] = true,
};

function Questo_API.getSpecialQuestInfo(questID)
	return SEAL_QUESTS[questID]
end

local SUPPORTED_SLOTS = {
	["INVTYPE_AMMO"]={"AmmoSlot"},
	["INVTYPE_HEAD"]={"HeadSlot"},
	["INVTYPE_NECK"]={"NeckSlot"},
	["INVTYPE_SHOULDER"]={"ShoulderSlot"},
	["INVTYPE_CHEST"]={"ChestSlot"},
	["INVTYPE_WAIST"]={"WaistSlot"},
	["INVTYPE_LEGS"]={"LegsSlot"},
	["INVTYPE_FEET"]={"FeetSlot"},
	["INVTYPE_WRIST"]={"WristSlot"},
	["INVTYPE_HAND"]={"HandsSlot"},
	["INVTYPE_FINGER"]={"Finger0Slot", "Finger1Slot"},
	["INVTYPE_TRINKET"]={"Trinket0Slot", "Trinket1Slot"},
	["INVTYPE_CLOAK"]={"BackSlot"},
	["INVTYPE_WEAPON"]={"MainHandSlot", "SecondaryHandSlot"},
	["INVTYPE_2HWEAPON"]={"MainHandSlot"},
	["INVTYPE_RANGED"]={"MainHandSlot"},
	["INVTYPE_RANGEDRIGHT"]={"MainHandSlot"},
	["INVTYPE_WEAPONMAINHAND"]={"MainHandSlot"},
	["INVTYPE_SHIELD"]={"SecondaryHandSlot"},
	["INVTYPE_WEAPONOFFHAND"]={"SecondaryHandSlot"},
	["INVTYPE_HOLDABLE"]={"SecondaryHandSlot"}
}

local function getItemLevel(itemLink)
	if (not itemLink) then
		return 0
	end
	local invQuality, invLevel = select(3, GetItemInfo(itemLink));
	return (invQuality == 7) and huge or invLevel;
end

local AUTO_EQUIP_DELAY = 2;
local function autoEquip(itemLink)

	if not Questo_Data.config.autoEquip then
		return
	end

	if not itemLink then return end;

	local name, link, quality, lootLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(itemLink);
	log(("autoEquip %s on slot %s"):format(name, equipSlot));

	local shouldAutoEquip = false;
	local equipOn;
	if Questo_Data.config.autoEquip then
		local slot = SUPPORTED_SLOTS[equipSlot]
		if slot then
			log(("Supported slot %s"):format(equipSlot));
			local firstSlot = GetInventorySlotInfo(slot[1]);
			local invLink = GetInventoryItemLink("player", firstSlot);
			local eqLevel = getItemLevel(invLink);
			
			if #slot > 1 then
				local secondSlot = GetInventorySlotInfo(slot[2]);
				invLink = GetInventoryItemLink("player", secondSlot);
				if invLink then
					local eq2Level = getItemLevel(invLink);
					firstSlot = (eqLevel > eq2Level) and secondSlot or firstSlot;
					eqLevel = (eqLevel > eq2Level) and eq2Level or eqLevel;
				end
			end

			log(("Comparing lvl: lootLevel %s vs eqLevel %s"):format(lootLevel, eqLevel));
			if lootLevel > eqLevel then
				shouldAutoEquip = true;
				equipOn = firstSlot;
			end
		end
	end

	if shouldAutoEquip then
		log(("Will auto equip %s on slot %s"):format(name, equipOn));
		after(AUTO_EQUIP_DELAY, function()
			if InCombatLockdown() then
				return;
			end
			for container=0, NUM_BAG_SLOTS do
				for slot=1, GetContainerNumSlots(container) do
					local link = GetContainerItemLink (container, slot);
					if link then
						local itemName = GetItemInfo(link);
						if itemName == name then
							log(("Found and trying to auto equip %s"):format(itemName, equipOn));
							EquipItemByName(name, equipOn);
							return;
						end
					end
				end
			end
		end);
	end
end
Questo_API.autoEquip = autoEquip;

local function autoEquipAllReward()
	if GetNumQuestRewards() > 0 then
		for i=1, GetNumQuestRewards() do
			local link = GetQuestItemLink("reward", i);
			autoEquip(link);
		end
	end
end
Questo_API.autoEquipAllReward = autoEquipAllReward;

local function getQuestData(qTitle)
	for questIndex=1, GetNumQuestLogEntries() do
		local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(questIndex);
		if questTitle == qTitle then
			SelectQuestLogEntry(questIndex);
			local questDescription, questObjectives = GetQuestLogQuestText();
			return questObjectives or "";
		end
	end
	return "";
end

local function showQuestPortraitFrame(isOnCompleteStep)
	if not Questo_Data.config.hideOriginalFrames then
		return;
	end

	local questPortrait, questPortraitText, questPortraitName, questPortraitMount

	if isOnCompleteStep then
		questPortrait, questPortraitText, questPortraitName, questPortraitMount = GetQuestPortraitTurnIn();
	else
		questPortrait, questPortraitText, questPortraitName, questPortraitMount = GetQuestPortraitGiver();
	end

	if questPortrait and questPortrait ~= 0 then
		QuestFrame_HideQuestPortrait(Questo_NPCFrame, questPortrait, questPortraitMount, questPortraitText, questPortraitName, -16, -48);
		QuestModelScene:SetFrameStrata("LOW")
	else
		QuestFrame_HideQuestPortrait();
	end
end

local function acceptQuest()
	if QuestFlagsPVP() then
		QuestFrame.dialog = StaticPopup_Show("CONFIRM_ACCEPT_PVP_QUEST");
	elseif QuestGetAutoAccept() then
		AcknowledgeAutoAcceptQuest();
		PlayAutoAcceptQuestSound();
		Questo_API.layout.hideQuestoFrame();
	else
		AcceptQuest();
		Questo_API.layout.hideQuestoFrame();
	end
end

local function gossipEventHandler(eventType)
	local dialogChoices = Questo_API.dialogs.getChoices(eventType);
	local buttonIndex = 0;
	local previousFrame;
	local totalButtonHeights = 0;

	for bucketType, bucket in pairs(dialogChoices) do
		for choiceIndex, choice in pairs(bucket) do
			buttonIndex = buttonIndex + 1;
			local dialogChoiceButton = DialogsButtons.getButton(DialogsScrollFrame.container, previousFrame);
			dialogChoiceButton:Decorate(buttonIndex, choiceIndex, choice, bucketType, eventType);

			totalButtonHeights = totalButtonHeights + dialogChoiceButton:GetHeight() + DialogsButtons.getMargin();
			previousFrame = dialogChoiceButton;
		end
	end

	if buttonIndex > 0 then
		DialogsScrollFrame.show(totalButtonHeights);
	end
end

eventHandlers[Dialogs.EVENT_TYPES.GOSSIP_SHOW] = function()
	gossipEventHandler(Dialogs.EVENT_TYPES.GOSSIP_SHOW);
	if Questo_Data.config.forceGossip and  not Questo_API.isCurrentNPCBlacklisted() then
		Questo_NPCFrameBlacklistButton:Show()
	end
end;
eventHandlers[Dialogs.EVENT_TYPES.QUEST_GREETING] = function()
	gossipEventHandler(Dialogs.EVENT_TYPES.QUEST_GREETING);
end

local HOVERED_FRAME_TITLE_MARGIN = 10;
local HOVERED_FRAME_TEXT_MARGIN = 5;

local function setObjectiveText(fontString, text, anchor)
	fontString:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, -1 * HOVERED_FRAME_TEXT_MARGIN);
	fontString:SetText(text);
	fontString:Show();
	return fontString:GetHeight() + HOVERED_FRAME_TEXT_MARGIN;
end

eventHandlers["QUEST_DETAIL"] = function()

	local previousText = Questo_NPCFrameObjectivesContent.Title;

	Questo_NPCFrameObjectivesContent.Title:SetPoint("TOP", 0, -1 * HOVERED_FRAME_TITLE_MARGIN);
	local contentHeight = Questo_NPCFrameObjectivesContent.Title:GetHeight() + HOVERED_FRAME_TITLE_MARGIN;

	Questo_NPCFrameObjectives:Show();
	Questo_NPCFrameObjectivesImage:SetDesaturated(false);

	local objectives = GetObjectiveText();
	if objectives:len() > 0 then
		contentHeight = contentHeight + setObjectiveText(Questo_NPCFrameObjectivesContent.Objectives, objectives, previousText);
		previousText = Questo_NPCFrameObjectivesContent.Objectives;
	end

	local groupNum = GetSuggestedGroupNum();
	if groupNum > 0 then
		contentHeight = contentHeight +  setObjectiveText(Questo_NPCFrameObjectivesContent.GroupSuggestion, format(QUEST_SUGGESTED_GROUP_NUM, groupNum), previousText);
		previousText = Questo_NPCFrameObjectivesContent.GroupSuggestion;
	end

	local rewardsBucket = Rewards.getRewards();

	for bucketType, bucket in pairs(rewardsBucket) do
		if tsize(bucket) > 0 then
			contentHeight = contentHeight + RewardsButtons.displayRewardsOnGrid(bucketType, bucket, Questo_NPCFrameObjectivesContent, previousText);
		end
	end

	contentHeight = contentHeight + HOVERED_FRAME_TEXT_MARGIN;

	Questo_NPCFrameObjectivesContent:SetHeight(contentHeight);

	if GetNumQuestItems() > 0 then
		local _, icon = GetQuestItemInfo("required", 1);
		Questo_NPCFrameObjectivesImage:SetTexture(icon);
	end
end

eventHandlers["QUEST_PROGRESS"] = function()

	local previousText = Questo_NPCFrameObjectivesContent.Title;

	Questo_NPCFrameObjectivesContent.Title:SetPoint("TOP", 0, -1 * HOVERED_FRAME_TITLE_MARGIN);
	local contentHeight = Questo_NPCFrameObjectivesContent.Title:GetHeight() + HOVERED_FRAME_TITLE_MARGIN;

	Questo_NPCFrameObjectives:Show();
	Questo_NPCFrameObjectivesImage:SetDesaturated(not IsQuestCompletable());
	local objectives = GetObjectiveText();
	if objectives:len() > 0 then
		if IsQuestCompletable() then
			objectives = getTextureString("Interface\\RAIDFRAME\\ReadyCheck-Ready", 15) .. " |cff00ff00" .. objectives;
		else
			objectives = getTextureString("Interface\\RAIDFRAME\\ReadyCheck-NotReady", 15) .. " |cffff0000" .. objectives;
		end
		contentHeight = contentHeight + setObjectiveText(Questo_NPCFrameObjectivesContent.Objectives, objectives, previousText);
		previousText = Questo_NPCFrameObjectivesContent.Objectives;
	end

	local groupNum = GetSuggestedGroupNum();
	if groupNum > 0 then
		contentHeight = contentHeight +  setObjectiveText(Questo_NPCFrameObjectivesContent.GroupSuggestion, format(QUEST_SUGGESTED_GROUP_NUM, groupNum), previousText);
		previousText = Questo_NPCFrameObjectivesContent.GroupSuggestion;
	end

	local objectivesBucket = Rewards.getObjectiveItems();

	if tsize(objectivesBucket) > 0 then
		contentHeight = contentHeight + RewardsButtons.displayRewardsOnGrid(Rewards.BUCKET_TYPES.OBJECTIVES, objectivesBucket, Questo_NPCFrameObjectivesContent, previousText);
	end

	contentHeight = contentHeight + HOVERED_FRAME_TEXT_MARGIN;

	Questo_NPCFrameObjectivesContent:SetHeight(contentHeight);
end

local CLICKING_ON_REWARDS_MEANS_CHOOSING_IT = true;
eventHandlers["QUEST_COMPLETE"] = function(eventInfo)

	local rewardsBucket, bestIcon, totalNumberOfRewards = Rewards.getRewards();

	if totalNumberOfRewards > 0 then

		Questo_NPCFrameRewards:Show();

		local previousText = Questo_NPCFrameRewards.Content.Title;

		Questo_NPCFrameRewards.Content.Title:SetPoint("TOP", 0, -1 * HOVERED_FRAME_TITLE_MARGIN);
		local contentHeight = Questo_NPCFrameRewards.Content.Title:GetHeight() + HOVERED_FRAME_TITLE_MARGIN;

		for bucketType, bucket in pairs(rewardsBucket) do
			if tsize(bucket) > 0 then
				contentHeight = contentHeight + RewardsButtons.displayRewardsOnGrid(bucketType, bucket, Questo_NPCFrameRewards.Content, previousText, CLICKING_ON_REWARDS_MEANS_CHOOSING_IT);
			end
		end

		contentHeight = contentHeight + HOVERED_FRAME_TEXT_MARGIN;

		Questo_NPCFrameObjectivesContent:SetHeight(contentHeight);

		Questo_NPCFrameRewardsItemIcon:SetTexture(bestIcon);
		Questo_NPCFrameRewards.Content:SetHeight(contentHeight);
	else
		Questo_NPCFrameRewards:Hide();
	end
end

local currentEvent;

function Questo_API.getCurrentEvent()
	return currentEvent;
end

local function handleEventSpecifics(event, texts, textIndex, eventInfo)

	Questo_NPCFrameGossipChoices:Hide();
	Questo_NPCFrameRewards:Hide();
	Questo_NPCFrameObjectives:Hide();
	Questo_NPCFrameObjectivesYes:Hide();
	Questo_NPCFrameObjectivesNo:Hide();
	Questo_NPCFrameObjectives.OK:Hide();
	Questo_NPCFrameObjectivesContent.RequiredItemText:Hide();
	Questo_NPCFrameObjectivesContent.GroupSuggestion:Hide();
	Questo_NPCFrameObjectivesContent.Objectives:SetText('');
	Questo_NPCFrameObjectivesContent.Objectives:Hide();
	Questo_NPCFrameRewards.Content:Hide();
	setTooltipForSameFrame(Questo_NPCFrameObjectives);
	Questo_NPCFrameObjectivesImage:SetTexture("Interface\\FriendsFrame\\FriendsFrameScrollIcon");
	QuestFrame_HideQuestPortrait();

	DialogsScrollFrame.hide();
	DialogsButtons.hideAllButtons();
	RewardsButtons.hideAllButtons();

	showQuestPortraitFrame(event == "QUEST_COMPLETE");

	if textIndex == #texts and eventHandlers[event] then
		currentEvent = event;
		eventHandlers[event](eventInfo);
	end
end

local EMOTE_COLOR;
local function playText(textIndex, targetModel)

	if not EMOTE_COLOR then
		EMOTE_COLOR = Ellyb.Color(ChatTypeInfo["MONSTER_EMOTE"]):Freeze()
	end

	local text = Questo_NPCFrameChat.texts[textIndex];
	local delay = GOSSIP_DELAY;
	local textLineToken = getId();

	Questo_NPCFrameChatText:SetTextColor(ChatTypeInfo["MONSTER_SAY"].r, ChatTypeInfo["MONSTER_SAY"].g, ChatTypeInfo["MONSTER_SAY"].b);

	local stillEmote = Questo_NPCFrameChat.stillEmote[textIndex];
	if text:byte() == 60 or not UnitExists("npc") or UnitIsUnit("player", "npc") or UnitIsDead("npc") or stillEmote then
		local colorCodeStart = EMOTE_COLOR:GetColorCodeStartSequence();
		local displayedText = text:gsub("|r", "|r" .. colorCodeStart)
		displayedText = displayedText:gsub("<", colorCodeStart .. "<");
		displayedText = displayedText:gsub(">", ">|r");

		if stillEmote then
			displayedText = EMOTE_COLOR(displayedText);
		end

		Questo_NPCFrameChatText:SetText(displayedText);
	else
		Questo_NPCFrameChatText:SetText(text);
		text:gsub("[%.%?%!]+", function(finder)
		end);
	end

	Questo_NPCFrameChat.start = 0;

	if #Questo_NPCFrameChat.texts > 1 then
		Questo_NPCFrameChatPrevious:Show();
	end

	handleEventSpecifics(Questo_NPCFrameChat.event, Questo_NPCFrameChat.texts, textIndex, Questo_NPCFrameChat.eventInfo);

	Questo_NPCFrameChat:SetHeight(Questo_NPCFrameChatText:GetHeight() + Questo_NPCFrameChatName:GetHeight() + Questo_NPCFrameChatNextText:GetHeight() + 50);
end

function Questo_API.playNext(targetModel)
	Questo_NPCFrameChatNext:Enable();
	Questo_NPCFrameChat.currentIndex = Questo_NPCFrameChat.currentIndex + 1;

	Questo_NPCFrameChatNextText:SetText(loc("Q_NEXT"));
	if Questo_NPCFrameChat.currentIndex >= #Questo_NPCFrameChat.texts then
		if Questo_NPCFrameChat.eventInfo.finishText and (type(Questo_NPCFrameChat.eventInfo.finishText) ~= "function" or Questo_NPCFrameChat.eventInfo.finishText()) then
			if type(Questo_NPCFrameChat.eventInfo.finishText) == "function" then
				Questo_NPCFrameChatNextText:SetText(Questo_NPCFrameChat.eventInfo.finishText());
			else
				Questo_NPCFrameChatNextText:SetText(Questo_NPCFrameChat.eventInfo.finishText);
			end
		end
	end

	if Questo_NPCFrameChat.currentIndex <= #Questo_NPCFrameChat.texts then
		playText(Questo_NPCFrameChat.currentIndex, targetModel);
	else
		if Questo_NPCFrameChat.eventInfo.finishMethod then
			Questo_NPCFrameChat.eventInfo.finishMethod();
		else
			hideQuestoFrame();
		end
	end
end

local targetModel = Questo_NPCFrame.models.you;
local playerModel = Questo_NPCFrame.models.me;
local ANIMATIONS = Questo_API.ANIMATIONS;

function Questo_API.initEventsStructure()
	local startDialog = Questo_API.startDialog;

	EVENT_INFO = {
		["QUEST_GREETING"] = {
			text = GetGreetingText,
			finishMethod = function()
				local firstChoice, bucketType, index = Dialogs.getFirstChoice(Dialogs.EVENT_TYPES.QUEST_GREETING);

				if Dialogs.getDialogChoiceSelectorForEventType(Dialogs.EVENT_TYPES.QUEST_GREETING, bucketType) then
					debug(("QUEST_GREETING – Finish method : Using selector method found for bucket type %s at index %s."):format(bucketType, index));
					Dialogs.getDialogChoiceSelectorForEventType(Dialogs.EVENT_TYPES.QUEST_GREETING, bucketType)(index);
				else
					debug("QUEST_GREETING – Finish method : No valid options found, fallback to closing quest.");
					CloseQuest();
				end
			end,
			finishText = function()
				local finishText = GOODBYE;
				local firstChoice = Dialogs.getFirstChoice(Dialogs.EVENT_TYPES.QUEST_GREETING);

				if firstChoice then
					finishText = firstChoice.title;
					debug(("QUEST_GREETING – Finish text : Found first choice with text %s."):format(finishText));
				else
					debug(("QUEST_GREETING – Finish text : Could not find a first choice, using default finish text %s."):format(finishText));
				end

				return finishText;
			end,
			cancelMethod = CloseQuest,
			titleGetter = GetTitleText,
		},
		["QUEST_DETAIL"] = {
			text = GetQuestText,
			cancelMethod = CloseQuest,
			titleGetter = GetTitleText,
			finishText = loc("Q_CHECK_OBJ"),
			finishMethod = function()
				if not Questo_NPCFrameObjectivesContent:IsVisible() then
					configureHoverFrame(Questo_NPCFrameObjectivesContent, Questo_NPCFrameObjectives, "TOP");
					setTooltipForSameFrame(Questo_NPCFrameObjectives, "TOP", 0, 0, nil, nil);
					Questo_MainTooltip:Hide();
					Questo_NPCFrameObjectivesYes:Show();
					Questo_NPCFrameObjectivesNo:Show();
				else
					acceptQuest();
				end
			end,
		},
		["QUEST_PROGRESS"] = {
			text = GetProgressText,
			finishMethod = function()
				if not Questo_NPCFrameObjectivesContent:IsVisible() then
					configureHoverFrame(Questo_NPCFrameObjectivesContent, Questo_NPCFrameObjectives, "TOP");
					setTooltipForSameFrame(Questo_NPCFrameObjectives, "TOP", 0, 0, nil, nil);
					Questo_MainTooltip:Hide();
					if IsQuestCompletable() then
						Questo_NPCFrameObjectives.OK:Show();
						Questo_NPCFrameChatNextText:SetText(loc("Q_CONTINUE"));
					else
						Questo_NPCFrameChatNextText:SetText(loc("Q_NOT_YET"));
					end
				elseif IsQuestCompletable() then
					CompleteQuest();
				else
					CloseQuest();
				end
			end,
			finishText = function()
				return loc("Q_CHECK_OBJ");
			end,
			cancelMethod = CloseQuest,
			titleGetter = GetTitleText,
		},
		["QUEST_COMPLETE"] = {
			text = GetRewardText,
			finishMethod = function()
				local rewardsBucket, bestIcon, totalNumberOfRewards = Rewards.getRewards();

				if not Questo_NPCFrameRewards.Content:IsVisible() and totalNumberOfRewards > 0 then
					configureHoverFrame(Questo_NPCFrameRewards.Content, Questo_NPCFrameRewardsItem, "TOP");
					setTooltipForSameFrame(Questo_NPCFrameRewardsItem, "TOP", 0, 0);
					Questo_MainTooltip:Hide();
					if GetNumQuestChoices() > 1 then
						Questo_NPCFrameChatNextText:SetText(loc("Q_SELECT_REWARD"));
						Questo_NPCFrameChatNext:Disable();
					else
						Questo_NPCFrameChatNextText:SetText(loc("Q_CONTINUE"));
					end

					local rewards = Rewards.getRewards();
					if  tsize(rewards[Rewards.BUCKET_TYPES.CHOICE]) > 0 then
					end

				elseif GetNumQuestChoices() == 1 then
					GetQuestReward(1);
					autoEquip(GetQuestItemLink("choice", 1));
					autoEquipAllReward();
				elseif GetNumQuestChoices() == 0 then
					GetQuestReward();
					autoEquipAllReward();
				end
			end,
			finishText = function()
				local rewardsBucket, bestIcon, totalNumberOfRewards = Rewards.getRewards();

				return totalNumberOfRewards > 0 and loc("Q_GET_REWARD") or Questo_NPCFrameChatNextText:SetText(loc("Q_CONTINUE"));
			end,
			cancelMethod = CloseQuest,
			titleGetter = GetTitleText,
		},
		["GOSSIP_SHOW"] = {
			text = GetGossipText,
			finishMethod = function()
				local firstChoice, bucketType, index = Dialogs.getFirstChoice(Dialogs.EVENT_TYPES.GOSSIP_SHOW);

				if firstChoice and Dialogs.getDialogChoiceSelectorForEventType(Dialogs.EVENT_TYPES.GOSSIP_SHOW, bucketType) then
					debug(("GOSSIP_SHOW – Finish method : Using selector method found for bucket type %s at index %s."):format(tostring(bucketType), tostring(index)));
					Dialogs.getDialogChoiceSelectorForEventType(Dialogs.EVENT_TYPES.GOSSIP_SHOW, bucketType)(index);
				else
					debug("GOSSIP_SHOW – Finish method : No valid options found, fallback to closing dialog.");
					CloseGossip();
				end
			end,
			finishText = function()
				local finishText = GOODBYE;
				local firstChoice = Dialogs.getFirstChoice(Dialogs.EVENT_TYPES.GOSSIP_SHOW);

				if firstChoice then
					finishText = firstChoice.title;
					debug(("GOSSIP_SHOW – Finish text : Found first choice with text %s."):format(finishText));
				else
					debug(("GOSSIP_SHOW – Finish text : Could not find a first choice, using default finish text %s."):format(finishText));
				end

				return finishText;
			end,
			cancelMethod = CloseGossip,
		},

	};
	Questo_API.EVENT_INFO = EVENT_INFO;

	local QuestoFrameShouldOpen = false;

	for event, info in pairs(EVENT_INFO) do
		Ellyb.GameEvents.registerCallback(event, function(...)

			if event == "QUEST_DETAIL" then
				local questStartItemID = ...;
				if (questStartItemID ~= nil and questStartItemID ~= 0) or (QuestGetAutoAccept() and QuestIsFromAreaTrigger()) or GetQuestID() == 0 then
					return
				end

			end
			if Questo_Data.config.disableInInstances then
				if IsInInstance() then
					return
				end
			end
			if Questo_Data.config.disableInDMF then
				local _, _, _, mapID = UnitPosition("player");
				if mapID and mapID == 974 then
					return
				end
			end

			if not ForceGossip() then
				QuestoFrameShouldOpen = true;
				C_Timer.After(0.5, function()
					if QuestoFrameShouldOpen then
						startDialog("npc", info.text(), event, info);
					end
				end)
			else
				startDialog("npc", info.text(), event, info);
			end
		end);
	end



	Ellyb.GameEvents.registerCallback("QUEST_ITEM_UPDATE", RewardsButtons.refreshButtons);
	Ellyb.GameEvents.registerCallback("GOSSIP_CLOSED", function()
		QuestoFrameShouldOpen = false;
	end);

	Questo_NPCFrameObjectivesYes:SetScript("OnClick", acceptQuest);
	Questo_NPCFrameObjectivesYes:SetScript("OnEnter", function(self)
		refreshTooltipForFrame(self);
	end);
	Questo_NPCFrameObjectivesNo:SetScript("OnClick", DeclineQuest);
	Questo_NPCFrameObjectivesNo:SetScript("OnEnter", function(self)
		refreshTooltipForFrame(self);
	end);

	Questo_NPCFrameObjectives:SetScript("OnClick", function() EVENT_INFO["QUEST_PROGRESS"].finishMethod(); end);
	Questo_NPCFrameObjectivesContent.Title:SetText(QUEST_OBJECTIVES);
	Questo_NPCFrameObjectivesContent.RequiredItemText:SetText(TURN_IN_ITEMS);

	Questo_NPCFrameRewardsItem:SetScript("OnClick", function() EVENT_INFO["QUEST_COMPLETE"].finishMethod(); end);
	Questo_NPCFrameRewards.Content.Title:SetText(REWARDS);

	hooksecurefunc("QuestInfo_ShowRewards", hideQuestRewardFrameIfNeed);

	local function goBackOnRightClick(_, button)
		if button == "RightButton" then
			Questo_API.goBackToPreviousStep();
		end
	end

	Questo_NPCFrameObjectivesContent:SetScript("OnMouseDown", goBackOnRightClick);
	Questo_NPCFrameRewards.Content:SetScript("OnMouseDown", goBackOnRightClick);
end
