local ShowUIPanel = ShowUIPanel;
local HideUIPanel = HideUIPanel;

local Questo_NPCFrame = Questo_NPCFrame;


local defaultFramesWeWantToReplace = {
	"QuestFrame",
	"GossipFrame",
};

Questo_API.layout = {};

local lockFrame = function(shouldLockTheFrame)
	if shouldLockTheFrame == nil then
		shouldLockTheFrame = true;
	end
	Questo_Data.config.lockFrame = shouldLockTheFrame;
end
Questo_API.layout.lockFrame = lockFrame;

local unlockFrame = function()
	Questo_Data.config.lockFrame = false;
end
Questo_API.layout.unlockFrame = unlockFrame;

local isFrameLocked = function()
	return Questo_Data.config.lockFrame == true;
end
Questo_API.layout.isFrameLocked = isFrameLocked;

local Questo_FrameLayoutOptions = {
	area = "left",
	pushable = 1,
	width = 830,
	height = 519
};

local registerToUILayoutEngine = function()

	UIPanelWindows["Questo_NPCFrame"] = Questo_FrameLayoutOptions;

	Questo_NPCFrame:SetSize(Questo_FrameLayoutOptions.width, Questo_FrameLayoutOptions.height);
	Questo_NPCFrameResizeButton.onResizeStop(Questo_FrameLayoutOptions.width, Questo_FrameLayoutOptions.height);

	lockFrame();

	Questo_NPCFrameLock:Hide();
	Questo_NPCFrameResizeButton:Hide();
end
Questo_API.layout.registerToUILayoutEngine = registerToUILayoutEngine;

local hiddenFrames = CreateFrame("FRAME");
hiddenFrames:Hide();

local unregisterFromUILayoutEngine = function()
	UIPanelWindows["Questo_NPCFrame"] = nil;
	Questo_NPCFrameLock:Show();
	Questo_NPCFrameResizeButton:Show();
end
Questo_API.layout.unregisterFromUILayoutEngine = unregisterFromUILayoutEngine;

local framesUILayoutEngineSettings = {};

local removeFrameFromUILayoutEngine = function(frameName)
	local info = UIPanelWindows[frameName];
	if not info then return end

	framesUILayoutEngineSettings[frameName] = info;

	local frame = _G[frameName];
	frame:SetAttribute("UIPanelLayout-defined", false);
	UIPanelWindows[frameName] = nil;
	frame:SetParent(hiddenFrames);
end

local addToLayoutEngine = function(frameName)
	local frame = _G[frameName];
	UIPanelWindows[frameName] = framesUILayoutEngineSettings[frameName];
	frame:SetAttribute("UIPanelLayout-defined", true);
	for name, value in pairs(UIPanelWindows[frameName]) do
		frame:SetAttribute("UIPanelLayout-" .. name, value);
	end
	frame:SetParent(UIParent);
end

Questo_API.layout.hideQuestRewardFrameIfNeed = function()
	if Questo_Data.config.hideOriginalFrames then
		QuestInfoRewardsFrame:Hide();
	else
		QuestInfoRewardsFrame:Show();
	end
end

local useUILayoutEngine = function()
	return Questo_Data.config.useLayoutEngine;
end

local showQuestoFrame = function()
	if useUILayoutEngine() then
		ShowUIPanel(Questo_NPCFrame);
	else
		Questo_NPCFrame:Show();
	end
	Questo_API.PlaySound("QUESTLOGOPEN");
end
Questo_API.layout.showQuestoFrame = showQuestoFrame;

local hideQuestoFrame = function()
	Questo_NPCFrameBlacklistButton:Hide()
	if useUILayoutEngine() then
		HideUIPanel(Questo_NPCFrame);
	else
		Questo_NPCFrame:Hide();
	end
	Questo_API.PlaySound("QUESTLOGCLOSE");
end
Questo_API.layout.hideQuestoFrame = hideQuestoFrame;

local toggleQuestoFrame = function(showFrame)
	if showFrame == nil then
		showFrame = not Questo_NPCFrame:IsShown();
	end
	if showFrame then
		showQuestoFrame();
	else
		hideQuestoFrame();
	end
end
Questo_API.layout.toggleQuestoFrame = toggleQuestoFrame;

Questo_API.layout.showDefaultFrames = function()
	hideQuestoFrame();
	for _, frame in pairs(defaultFramesWeWantToReplace) do
		addToLayoutEngine(frame);
	end
end

local hideDefaultFrames = function()
	hideQuestoFrame();
	for _, frame in pairs(defaultFramesWeWantToReplace) do
		_G[frame]:Hide();
		removeFrameFromUILayoutEngine(frame)
	end
end
Questo_API.layout.hideDefaultFrames = hideDefaultFrames;

Ellyb.GameEvents.registerCallback("PLAYER_ENTERING_WORLD", function()
	if Questo_Data.config.hideOriginalFrames then
		Questo_API.layout.hideDefaultFrames();
		if Questo_Data.config.disableInInstances then
			if IsInInstance() then
				Questo_API.layout.showDefaultFrames();
			end
		end
		if Questo_Data.config.disableInDMF then
			local _, _, _, mapID = UnitPosition("player");
			if mapID and mapID == 974 then
				Questo_API.layout.showDefaultFrames();
			end
		end
	end
end)


Questo_NPCFrame:SetScript("OnHide", function()
	if Questo_NPCFrameChat.eventInfo and Questo_NPCFrameChat.eventInfo.cancelMethod then
		Questo_NPCFrameChat.eventInfo.cancelMethod();
		Questo_NPCFrameChat.eventInfo = nil;
	end
	if Questo_Data.config.hideOriginalFrames then
		CloseGossip();
	end
end)
