local Ellyb = Ellyb(...);

local wipe, tContains = wipe, tContains;
local UnitGUID = UnitGUID;
local setTooltipForSameFrame, setTooltipAll = Questo_API.lib.setTooltipForSameFrame, Questo_API.lib.setTooltipAll;
local loc, tsize = Questo_API.locale.getText, Questo_API.lib.tsize;
local playNext = Questo_API.playNext;
local showQuestoFrame = Questo_API.layout.showQuestoFrame;
local hideQuestoFrame = Questo_API.layout.hideQuestoFrame;
local strtrim = strtrim;
local insert = table.insert;

local strsplit, pairs, tostring = strsplit, pairs, tostring;
local UnitIsUnit, UnitExists, UnitName = UnitIsUnit, UnitExists, UnitName;
local IsAltKeyDown, IsShiftKeyDown, IsControlKeyDown = IsAltKeyDown, IsShiftKeyDown, IsControlKeyDown;
local InterfaceOptionsFrame_OpenToCategory = InterfaceOptionsFrame_OpenToCategory;

local mainFrame = Questo_NPCFrame;
local targetModel = mainFrame.models.you;
targetModel.isModelDisplayedOnLeft = false;
local playerModel = mainFrame.models.me;

local scalingDB, customHeightDB, customPersonalDB;

local LINE_FEED_CODE = string.char(10);
local CARRIAGE_RETURN_CODE = string.char(13);
local WEIRD_LINE_BREAK = LINE_FEED_CODE .. CARRIAGE_RETURN_CODE .. LINE_FEED_CODE;
local CHAT_MARGIN = 70;

local function getScalingStuctures(modelMeID, modelYouID)
	local key, invertedKey = scalingLib:GetModelKeys(modelMeID, modelYouID);
	local dataMe, dataYou = scalingLib:GetModelCoupleProperties(modelMeID, modelYouID);

	if customHeightDB[key] then
		dataMe.scale = customHeightDB[key][1];
		dataYou.scale = customHeightDB[key][2];
	elseif customHeightDB[invertedKey] then
		dataMe.scale = customHeightDB[invertedKey][2];
		dataYou.scale = customHeightDB[invertedKey][1];
	end

	if customPersonalDB[modelMeID] then
		for field, value in pairs(customPersonalDB[modelMeID]) do
			dataMe[field] = value;
		end
	end
	if customPersonalDB[modelYouID] then
		for field, value in pairs(customPersonalDB[modelYouID]) do
			dataYou[field] = value;
		end
	end

	return dataMe, dataYou;
end

local function resetStructure()
	local key, invertedKey = scalingLib:GetModelKeys(playerModel:GetModelFileIDAsString(), targetModel:GetModelFileIDAsString());

	for _, value in pairs({key, invertedKey}) do
		if customHeightDB[value] then
			wipe(customHeightDB[value]);
			customHeightDB[value] = nil;
		end
	end

	if customPersonalDB[playerModel:GetModelFileIDAsString()] then
		wipe(customPersonalDB[playerModel:GetModelFileIDAsString()]);
		customPersonalDB[playerModel:GetModelFileIDAsString()] = nil;
	end
	if customPersonalDB[targetModel:GetModelFileIDAsString()] then
		wipe(customPersonalDB[targetModel:GetModelFileIDAsString()]);
		customPersonalDB[targetModel:GetModelFileIDAsString()] = nil;
	end
end

local Addon = CreateFrame("FRAME");
local addonName, addonTable = ...;
local L = addonTable["Locale"];
local letterBox;
local blizzardFont;
local timer = CreateFrame("FRAME");
local function createTimer(after, func)
	local total = 0;
	timer:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed;
		if(total > after) then
			self:SetScript("OnUpdate", nil);
			func();
		end
	end);
end

local frameFader = CreateFrame("FRAME");
local function hideLetterBox()
	if(not letterBox:IsShown()) then
		return;
	end
local alpha = letterBox:GetAlpha();
	letterBox:SetScript("OnUpdate", function(self, elapsed)
		letterBox:SetAlpha(0);
	end);
end

local function showLetterBox()
	if(IsModifierKeyDown() or not QuestFrame:IsShown()) then
		return;
	end
	
	frameFader:SetScript("OnUpdate", nil);
	local alpha = letterBox:GetAlpha();
	letterBox:Show();
	letterBox:SetScript("OnUpdate", function(self, elapsed)
		letterBox:SetAlpha(1);
	end);
end
	
local function setUpLetterBox()
	letterBox = QuestC;
	letterBox:SetAllPoints();
	letterBox:SetScript("OnHide", function(self)
	QCQuestTitleFrame2:Hide();
	end);
	letterBox:Hide();
end

local localeFonts = {
	["zhCN"] = "Fonts\\ARKai_T.TTF",
	["ruRU"] = "Fonts\\FRIZQT___CYR.TTF",
	["zhTW"] = "Fonts\\bLEI00D.TTF",
	["koKR"] = "Fonts\\2002.TTF",
}

local function onPlayerLogin()
	blizzardFont = localeFonts[GetLocale()] or STANDARD_TEXT_FONT;
	setUpLetterBox();
	Addon:RegisterEvent("GOSSIP_SHOW");
	Addon:RegisterEvent("MERCHANT_SHOW");
	Addon:RegisterEvent("TRAINER_SHOW");
	Addon:RegisterEvent("TAXIMAP_OPENED");
	Addon:RegisterEvent("GOSSIP_CLOSED");
	Addon:RegisterEvent("MERCHANT_CLOSED");
	Addon:RegisterEvent("TRAINER_CLOSED");
	Addon:RegisterEvent("TAXIMAP_CLOSED");
	Addon:RegisterEvent("QUEST_GREETING");
	Addon:RegisterEvent("QUEST_DETAIL");
    Addon:RegisterEvent("QUEST_PROGRESS");
	Addon:RegisterEvent("QUEST_COMPLETE");
	Addon:RegisterEvent("QUEST_FINISHED");
	Addon:RegisterEvent("QUEST_TURNED_IN");
end

local function onQuestDetail(questStartItemID)
	if (questStartItemID and questStartItemID ~= 0) then
		return;
	end
	 QCQuestTitleFrame2.levelFrame.questText:SetText(GetTitleText());
	 QCQuestTitleFrame2:Show();
	 showLetterBox();
	end


local function onQuestProgress()
	QCQuestTitleFrame2.levelFrame.questText:SetText(GetTitleText());
	QCQuestTitleFrame2:Show();
	showLetterBox();
    end

local function onQuestComplete()
	if(not letterBox:IsShown()) then
	QCQuestTitleFrame2.levelFrame.questText:SetText(GetTitleText());
	QCQuestTitleFrame2:Show();
	showLetterBox();
	end
	end

Addon.scripts = {
	["PLAYER_LOGIN"] = onPlayerLogin,
	["QUEST_DETAIL"] = onQuestDetail,
	["QUEST_PROGRESS"] = onQuestProgress,
	["QUEST_COMPLETE"] = onQuestComplete,
	["QUEST_TURNED_IN"] = onQuestTurnedIn,
};

Addon:SetScript("OnEvent", function(self, event, ...)
	if(Addon.scripts[event]) then
		Addon.scripts[event](...);
	elseif(event == "GOSSIP_CLOSED" or event == "MERCHANT_CLOSED" or event == "TRAINER_CLOSED" or event == "QUEST_FINISHED" or event == "TAXIMAP_CLOSED") then
		createTimer(0.05, function()
			hideLetterBox();
			SetView(-1);
		end);
	elseif(event == "MERCHANT_SHOW" or event == "TRAINER_SHOW" or event == "TAXIMAP_OPENED") then
	end
end);

Addon:RegisterEvent("PLAYER_LOGIN");

Questo_API.getQuestLevelColor = getQuestLevelColor;

local function modelsLoaded()
	playerModel:ResetIdleAnimationID();
	targetModel:ResetIdleAnimationID();

	local dataMe, dataYou = getScalingStuctures(playerModel:GetModelFileIDAsString(), targetModel:GetModelFileIDAsString());

	if playerModel.doNotAnimateScaling then
		playerModel:SetScalingValuesIn(dataMe.scale, dataMe.feet, dataMe.offset, dataMe.facing);
		playerModel.doNotAnimateScaling = false;
	else
		playerModel:AnimateScalingValuesIn(dataMe.scale, dataMe.feet, dataMe.offset, dataMe.facing);
	end

	if targetModel:GetModelFileIDAsString() then
		if targetModel.doNotAnimateScaling then
			targetModel:SetScalingValuesIn(dataYou.scale, dataYou.feet, dataYou.offset, dataYou.facing);
			targetModel.doNotAnimateScaling = false;
		else
			targetModel:AnimateScalingValuesIn(dataYou.scale, dataYou.feet, dataYou.offset, dataYou.facing);
		end
		Questo_NPCFrameChat.bubbleTail:Show();
	else
		playerModel:SetCustomIdleAnimationID(Questo_API.ANIMATIONS.READING);
		playerModel:ApplySpellVisualKit(29521, false)
		Questo_NPCFrameChat.bubbleTail:Hide();
		playerModel:PlayIdleAnimation();
	end
	if targetModel:GetModelFileIDAsString() then
		mainFrame.debug.you:SetText(targetModel:GetModelFileIDAsString());
	end
	if playerModel:GetModelFileIDAsString() then
		mainFrame.debug.me:SetText(playerModel:GetModelFileIDAsString());
	end

	mainFrame.debug.recorded:Hide();
	if scalingLib:IsRecorded(playerModel:GetModelFileIDAsString(), targetModel:GetModelFileIDAsString()) then
		mainFrame.debug.recorded:Show();
	end
end
Questo_API.onModelsLoaded = modelsLoaded;
playerModel.ModelLoaded = modelsLoaded;
targetModel.ModelLoaded = modelsLoaded;

local SLIDE_IN = 0.2;
local function setModelsAlpha(value)
	local slideInValue = SLIDE_IN - (SLIDE_IN * value);
	targetModel:SetAlpha(value);
	playerModel:SetAlpha(value);
	playerModel:SetPosition(nil, slideInValue * -1)
	targetModel:SetPosition(nil, slideInValue);
end

local alphaTransitionator = Ellyb.Transitionator();
local function animateInModels()
	alphaTransitionator:RunValue(0, 1, 0.5, setModelsAlpha)
end

local background = mainFrame.Background

function Questo_API.startDialog(targetType, fullText, event, eventInfo)

	local questId = GetQuestID()

	mainFrame.models.you.npc_id = Questo_API.getNpcId();

	if Questo_API.getNpcId(questId) then
		mainFrame.chat.name:SetText(UnitName("NPC"));
	else
		local targetName = UnitName(targetType) or ""
		if (not targetName or targetName:len() > 0 or targetName ~= UNKNOWN) and eventInfo.nameGetter and eventInfo.nameGetter() then
			targetName = eventInfo.nameGetter()
		end
		mainFrame.chat.name:SetText(targetName)
		mainFrame.chat.name:Show()
	end


	if eventInfo.titleGetter and eventInfo.titleGetter() and eventInfo.titleGetter():len() > 0 then
		mainFrame.Banner:Show();

		if C_CampaignInfo.IsCampaignQuest(questId) then
			local faction = UnitFactionGroup("player");
			if faction == "Horde" then
				mainFrame.Banner.FactionIcon:SetAtlas("bfa-landingbutton-horde-up")
			else
				mainFrame.Banner.FactionIcon:SetAtlas("bfa-landingbutton-alliance-up")
			end
			mainFrame.Banner.FactionIcon:Show()
		else
			mainFrame.Banner.FactionIcon:Hide()
		end

		mainFrame.Banner.Title:SetText(eventInfo.titleGetter());
		if eventInfo.getTitleColor and eventInfo.getTitleColor() then
			mainFrame.Banner.Title:SetTextColor(eventInfo.getTitleColor());
		else
			mainFrame.Banner.Title:SetTextColor(0.95, 0.95, 0.95);
		end
	else
		mainFrame.Banner.FactionIcon:Hide()
		mainFrame.Banner:Hide();
	end



	fullText = fullText:gsub(LINE_FEED_CODE .. "+", "\n");
	fullText = fullText:gsub(WEIRD_LINE_BREAK, "\n");

	local texts = {};
	for _, text in pairs({ strsplit("\n", fullText) }) do
		if strtrim(text) ~= "" then
			insert(texts, text);
		end
	end

	if texts[#texts] and texts[#texts]:len() == 0 then
		texts[#texts] = nil;
	end
	local stillEmote = { false };
	for index, text in pairs(texts) do
		if index < #texts then
			local prevEmote = stillEmote[index];
			local currentEmote = prevEmote;

			local _, openEmoteCount = text:gsub("<", "<");
			local _, closeEmoteCount = text:gsub(">", ">");

			if prevEmote and openEmoteCount < closeEmoteCount then
				currentEmote = false;
			elseif not prevEmote and openEmoteCount > closeEmoteCount then
				currentEmote = true;
			end

			stillEmote[index + 1] = currentEmote;
		end
	end

	mainFrame.chat.texts = texts;
	mainFrame.chat.currentIndex = 0;
	mainFrame.chat.eventInfo = eventInfo;
	mainFrame.chat.event = event;
	mainFrame.chat.stillEmote = stillEmote;
	Questo_NPCFrameObjectivesContent:Hide();
	mainFrame.chat.previous:Hide();

	if not mainFrame:IsVisible() then
		playerModel.doNotAnimateScaling = true;
		targetModel.doNotAnimateScaling = true;
		showQuestoFrame();
	end

	playNext(mainFrame.models.you);
end

local ANIMATION_TEXT_SPEED = 80;

local function onUpdateChatText(self, elapsed)
	if self.start and mainFrame.chat.text:GetText() and mainFrame.chat.text:GetText():len() > 0 then
		self.start = self.start + (elapsed * (ANIMATION_TEXT_SPEED * Questo_Data.config.textSpeedFactor or 0.5));
		if Questo_Data.config.textSpeedFactor == 0 or self.start >= mainFrame.chat.text:GetText():len() then
			self.start = nil;
			mainFrame.chat.text:SetAlphaGradient(mainFrame.chat.text:GetText():len(), 1);
		else
			mainFrame.chat.text:SetAlphaGradient(self.start, 30);
		end
	end
end

local function closeDialog()
	if mainFrame.chat.eventInfo and mainFrame.chat.eventInfo.cancelMethod then
		mainFrame.chat.eventInfo.cancelMethod();
	end
	hideQuestoFrame();
end

local function resetDialog()
	Questo_NPCFrameObjectivesContent:Hide();
	mainFrame.chat.currentIndex = 0;
	playNext(Questo_NPCFrameModelsYou);
end

function Questo_API.goBackToPreviousStep()
	Questo_NPCFrameObjectivesContent:Hide();
	mainFrame.chat.currentIndex = mainFrame.chat.currentIndex - 2;
	playNext(Questo_NPCFrameModelsYou);
end

Questo_API.addon = LibStub("AceAddon-3.0"):NewAddon("Questo", "AceConsole-3.0");

function Questo_API.getNpcId()
	local npcId = select(6, strsplit("-", UnitGUID("npc") or ""));
	return npcId;
end

function Questo_API.isCurrentNPCBlacklisted()
	return Questo_Data.npc_blacklist[Questo_API.getNpcId()] ~= nil
end

function Questo_API.addon:OnEnable()

	if not Questo_Data then
		Questo_Data = {};
	end

	if not Questo_Data.customscale then
		Questo_Data.customscale = {};
	end
	if not Questo_Data.customscale.relative then
		Questo_Data.customscale.relative = {};
	end
	if not Questo_Data.customscale.personal then
		Questo_Data.customscale.personal = {};
	end
	scalingDB = Questo_Data.customscale;
	customHeightDB = Questo_Data.customscale.relative;
	customPersonalDB = Questo_Data.customscale.personal;

	if not Questo_Data.config then
		Questo_Data.config = {};
	end
	if not Questo_Data.npc_blacklist then
		Questo_Data.npc_blacklist = {};
	end

	local NPC_IDS_WITH_BROKEN_DIALOGS = { }

	ForceGossip = function()
		local npcId = Questo_API.getNpcId();
		return Questo_Data.config.forceGossip and not NPC_IDS_WITH_BROKEN_DIALOGS[npcId] and not Questo_API.isCurrentNPCBlacklisted();
	end

	Questo_API.locale.init();

	mainFrame.chat.next:RegisterForClicks("LeftButtonUp", "RightButtonUp", "MiddleButtonUp");
	mainFrame.chat.next:SetScript("OnClick", function(self, button)
		if button == "RightButton" then
			if mainFrame.chat.currentIndex < #mainFrame.chat.texts then
				mainFrame.chat.currentIndex = #mainFrame.chat.texts - 1;
				playNext(mainFrame.models.you);
			else
				playNext(mainFrame.models.you);
			end
		elseif button == "MiddleButton" then
			closeDialog();
		else
			if mainFrame.chat.start and mainFrame.chat.start < mainFrame.chat.text:GetText():len() then
				mainFrame.chat.start = mainFrame.chat.text:GetText():len();
			else
				playNext(mainFrame.models.you);
			end
		end
	end);
	mainFrame.chat.previous:SetScript("OnClick", resetDialog);
	mainFrame.chat:SetScript("OnUpdate", onUpdateChatText);
	Questo_NPCFrameClose:SetScript("OnClick", closeDialog);
	Questo_NPCFrameRewardsItem:SetScale(1.5);

	Questo_NPCFrameBlacklistButton:SetScript("OnClick", function()
		Questo_Data.npc_blacklist[Questo_API.getNpcId()] = true;
		SelectGossipOption(1);
	end)
	Ellyb.Tooltips.getTooltip(Questo_NPCFrameBlacklistButton)
		:SetAnchor(Ellyb.Tooltips.ANCHORS.BOTTOMLEFT)
		:SetOffset(10, 10)

	mainFrame:SetScript("OnKeyDown", function(self, key)
		if not Questo_Data.config.useKeyboard then
			self:SetPropagateKeyboardInput(true);
			return;
		end

		if key == "SPACE" then
			self:SetPropagateKeyboardInput(false);
			mainFrame.chat.next:Click(IsShiftKeyDown() and "RightButton" or "LeftButton");
		elseif key == "BACKSPACE" then
			self:SetPropagateKeyboardInput(false);
			mainFrame.chat.previous:Click();
		elseif key == "ESCAPE" then
			closeDialog();
		else
			local keyNumber = tonumber(key);
			if not keyNumber or not Questo_API.dialogs.buttons.selectOptionAtIndex(keyNumber) then
				self:SetPropagateKeyboardInput(true);
			end
		end
	end);

	Questo_NPCFrameGossipChoices:SetScript("OnKeyDown", function(self, key)
		if not Questo_Data.config.useKeyboard then
			self:SetPropagateKeyboardInput(true);
			return;
		end

		if key == "ESCAPE" then
			Questo_NPCFrameGossipChoices:Hide();
			self:SetPropagateKeyboardInput(false);
			return;
		end

		local keyNumber = tonumber(key);
		if not keyNumber then
			self:SetPropagateKeyboardInput(true);
			return;
		end

		if keyNumber == 0 then
			keyNumber = 10;
		end

		local foundFrames = 0;
		for i = 0, 9 do
			if _G["Questo_ChoiceString" .. i] and _G["Questo_ChoiceString" .. i].IsVisible and _G["Questo_ChoiceString" .. i]:IsVisible() then
				foundFrames = foundFrames + 1;
				if foundFrames == keyNumber then
					_G["Questo_ChoiceString" .. i]:Click();
					self:SetPropagateKeyboardInput(false);
					return;
				end
			end
		end

		self:SetPropagateKeyboardInput(true);
		return;

	end);

	mainFrame.models.you.animTab = {};
	mainFrame.models.me.animTab = {};

	mainFrame.models.you:SetScript("OnUpdate", function(self, elapsed)
		if self.spin then
			self.spinAngle = self.spinAngle - (elapsed / 2);
			self:SetFacing(self.spinAngle);
		end
	end);

	Questo_API.initEventsStructure();

	Ellyb.GameEvents.registerCallback("GOSSIP_CLOSED", function()
		hideQuestoFrame();
	end);
	Ellyb.GameEvents.registerCallback("QUEST_FINISHED", function()
		hideQuestoFrame();
	end);

	local resizeChat = function()
		mainFrame.chat.text:SetWidth(mainFrame:GetWidth() - 150);
		mainFrame.chat:SetHeight(mainFrame.chat.text:GetHeight() + CHAT_MARGIN + 5);
		Questo_NPCFrameGossipChoices:SetWidth(mainFrame:GetWidth() - 400);
	end
	mainFrame.chat.text:SetWidth(550);
	Questo_NPCFrameResizeButton.onResizeStop = function(width, height)
		resizeChat();
		Questo_Data.config.width = width;
		Questo_Data.config.height = height;
		Questo_API.dialogs.scrollFrame.refreshMargins(width, height);
	end;
	mainFrame:SetSize(Questo_Data.config.width or 700, Questo_Data.config.height or 450);
	Questo_NPCFrameResizeButton.onResizeStop(Questo_Data.config.width or 700, Questo_Data.config.height or 450);
	resizeChat();

	Questo_API.addon:RegisterChatCommand("Questo", function()
		InterfaceOptionsFrame_OpenToCategory(QuestoOptionsPanel);
		if not Questo_NPCFrameConfigButton.shown then -- Dirty fix for the Interface frame shitting itself the first time
			Questo_NPCFrameConfigButton.shown = true;
			InterfaceOptionsFrame_OpenToCategory(QuestoOptionsPanel);
		end;
	end);

	setTooltipAll(Questo_NPCFrameConfigButton, "TOP", 0, 0, loc("Q_CONFIG"));


	mainFrame:RegisterForDrag("LeftButton");

	mainFrame:SetScript("OnDragStart", function(self)
		if not Questo_API.layout.isFrameLocked() then
			self:StartMoving();
		end
	end);

	mainFrame:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing();
	end);
	
	Questo_API.options.init();

end