local Ellyb = Ellyb(...);
local loc = Questo_API.locale.getText;
local setupListBox = Questo_API.lib.setupListBox;
local setTooltipForSameFrame = Questo_API.lib.setTooltipForSameFrame;
local QuestoOptionsPanel = QuestoOptionsPanel;
local QuestoTextOptionsPanel = QuestoTextOptionsPanel;
local QuestoMiscellaneousOptionsPanel = QuestoMiscellaneousOptionsPanel;
local Questo_NPCFrameChatName, Questo_NPCFrameChatText, Questo_NPCFrameTitle, Questo_NPCFrameChatNextText = Questo_NPCFrameChatName, Questo_NPCFrameChatText, Questo_NPCFrameTitle, Questo_NPCFrameChatNextText;

Questo_API.options = {}

local function hideOriginalFrames()
	GossipFrame:ClearAllPoints();
	local frameWidth = GossipFrame:GetWidth();
	GossipFrame:SetPoint("TOPLEFT", frameWidth * -1, 0);
	QuestFrame:ClearAllPoints();
	frameWidth = QuestFrame:GetWidth();
	QuestFrame:SetPoint("TOPLEFT", frameWidth * -1, 0);
end

Questo_API.options.hideOriginalFrames = hideOriginalFrames;

local function showOriginalFrames()
	GossipFrame:ClearAllPoints();
	GossipFrame:SetPoint("TOPLEFT", 16, -116);
	QuestFrame:ClearAllPoints();
	QuestFrame:SetPoint("TOPLEFT", 16, -116);
end

Questo_API.options.showOriginalFrames = showOriginalFrames;

local function decorateTextOptions(title, optionKey, affectedText)

	if not Questo_Data.config[optionKey] then
		Questo_Data.config[optionKey] = {}
	end

	QuestoTextOptionsPanel[optionKey].Title:SetText(title);
	QuestoTextOptionsPanel[optionKey].TextSample:SetText(loc("Q_CONFIG_SAMPLE_TEXT"));

	_G["QuestoTextOptionsPanel" .. optionKey .. "SizeSliderLow"]:SetText(9);
	_G["QuestoTextOptionsPanel" .. optionKey .. "SizeSliderHigh"]:SetText(25);
	QuestoTextOptionsPanel[optionKey].SizeSlider:SetScript("OnValueChanged", function(self, scale)
		_G["QuestoTextOptionsPanel" .. optionKey .. "SizeSliderText"]:SetText(scale);
		local font, _, outline = affectedText:GetFont();
		affectedText:SetFont(font, scale, outline);
		QuestoTextOptionsPanel[optionKey].TextSample:SetFont(font, scale, outline);
		Questo_Data.config[optionKey].Size = scale;
		Questo_NPCFrameChat:SetHeight(Questo_NPCFrameChatText:GetHeight() + Questo_NPCFrameChatName:GetHeight() + Questo_NPCFrameChatNextText:GetHeight() + 50);
	end);
	QuestoTextOptionsPanel[optionKey].SizeSlider:SetValue(Questo_Data.config[optionKey].Size or select(2, affectedText:GetFont()));

	local fonts = Questo_API.lib.getFonts();

	if not Questo_Data.config[optionKey].Font then
		Questo_Data.config[optionKey].Font = Questo_API.lib.getDefaultFont();
	end

	local font = Questo_API.lib.getFontPath(Questo_Data.config[optionKey].Font);
	local _, scale, outline = affectedText:GetFont();
	affectedText:SetFont(font, scale, outline);
	QuestoTextOptionsPanel[optionKey].TextSample:SetFont(font, scale, outline);

	setupListBox(QuestoTextOptionsPanel[optionKey].FontDropDown, fonts, function(fontIndex)
		local font = Questo_API.lib.getFontPath(fontIndex);
		local _, scale, outline = affectedText:GetFont();
		affectedText:SetFont(font, scale, outline);
		QuestoTextOptionsPanel[optionKey].TextSample:SetFont(font, scale, outline);
		Questo_Data.config[optionKey].Font = fontIndex;
	end, nil, 100, true, true);
	QuestoTextOptionsPanel[optionKey].FontDropDown:SetSelectedValue(Questo_Data.config[optionKey].Font);
end

local function decorateTextTemplateOptions(title, optionKey, defaultTextOptions, callback)

	if not Questo_Data.config[optionKey] then
		Questo_Data.config[optionKey] = {
			Font = Questo_API.lib.getDefaultFont(),
			Size = defaultTextOptions[2],
			Outline = defaultTextOptions[3]
		}
	end

	QuestoTextOptionsPanel[optionKey].Title:SetText(title);
	QuestoTextOptionsPanel[optionKey].TextSample:SetText(loc("Q_CONFIG_SAMPLE_TEXT"));

	_G["QuestoTextOptionsPanel" .. optionKey .. "SizeSliderLow"]:SetText(9);
	_G["QuestoTextOptionsPanel" .. optionKey .. "SizeSliderHigh"]:SetText(25);
	QuestoTextOptionsPanel[optionKey].SizeSlider:SetScript("OnValueChanged", function(self, scale)
		_G["QuestoTextOptionsPanel" .. optionKey .. "SizeSliderText"]:SetText(scale);
		Questo_Data.config[optionKey].Size = scale;
		QuestoTextOptionsPanel[optionKey].TextSample:SetFont(Questo_Data.config[optionKey].Font, Questo_Data.config[optionKey].Size, Questo_Data.config[optionKey].Outline);
		callback(Questo_Data.config[optionKey]);
	end);
	QuestoTextOptionsPanel[optionKey].SizeSlider:SetValue(Questo_Data.config[optionKey].Size);

	local fonts = Questo_API.lib.getFonts();
	local font = Questo_API.lib.getFontPath(Questo_Data.config[optionKey].Font);
	QuestoTextOptionsPanel[optionKey].TextSample:SetFont(font, Questo_Data.config[optionKey].Size, Questo_Data.config[optionKey].Outline);

	setupListBox(QuestoTextOptionsPanel[optionKey].FontDropDown, fonts, function(fontIndex)
		Questo_Data.config[optionKey].Font = fontIndex;
		local font = Questo_API.lib.getFontPath(fontIndex);
		QuestoTextOptionsPanel[optionKey].TextSample:SetFont(font, Questo_Data.config[optionKey].Size, Questo_Data.config[optionKey].Outline);
		callback(Questo_Data.config[optionKey]);
	end, nil, 100, true, true);
	QuestoTextOptionsPanel[optionKey].FontDropDown:SetSelectedValue(Questo_Data.config[optionKey].Font);

	callback(Questo_Data.config[optionKey]);
end

Questo_API.options.init = function()

	QuestoOptionsPanel.Title:SetText(loc("Q_CONFIG"));
	QuestoOptionsPanel.SubText:SetText(loc("Q_CONFIG_WELCOME"));


	local localeTab = {};
	for locale, localeInfo in pairs(Questo_API.locale.getLocales()) do
		tinsert(localeTab, {localeInfo.localeText, locale});
	end

	local init = true;
	QuestoOptionsPanel.Locale.Label:SetText(loc("Q_CONFIG_LANGUAGE"));
	setupListBox(QuestoOptionsPanel.Locale.DropDown, localeTab, function(locale)
		Questo_Data.config.locale = locale;
		if not init then
			ReloadUI();
		end
	end, nil, 100, true);
	QuestoOptionsPanel.Locale.DropDown:SetSelectedValue(Questo_Data.config.locale or Questo_API.locale.DEFAULT_LOCALE);
	init = false;

	QuestoOptionsPanel.ForceGossip.Text:SetText(loc("Q_CONFIG_FORCEGOSSIP"));
	QuestoOptionsPanel.ForceGossip.tooltip = loc("Q_CONFIG_FORCEGOSSIP_TT");
	QuestoOptionsPanel.ForceGossip:SetScript("OnClick", function(self)
		Questo_Data.config.forceGossip = self:GetChecked() == true;
	end);
	if Questo_Data.config.forceGossip == nil then
		Questo_Data.config.forceGossip = false;
	end
	QuestoOptionsPanel.ForceGossip:SetChecked(Questo_Data.config.forceGossip);

	QuestoOptionsPanel.HideOriginalFrames:SetScript("OnClick", function(self)
		Questo_Data.config.hideOriginalFrames = self:GetChecked() == true;
		if Questo_Data.config.hideOriginalFrames then
			Questo_API.layout.hideDefaultFrames();
		else
			Questo_API.layout.showDefaultFrames();
		end
	end);
	if Questo_Data.config.hideOriginalFrames == nil then
		Questo_Data.config.hideOriginalFrames = true;
	end
	QuestoOptionsPanel.HideOriginalFrames:SetChecked(Questo_Data.config.hideOriginalFrames);
	if Questo_Data.config.hideOriginalFrames then
		Questo_API.layout.hideDefaultFrames();
	end
	
	QuestoOptionsPanel.LockFrame:SetScript("OnClick", function(self)
		Questo_Data.config.lockFrame = self:GetChecked() == true;
		Questo_NPCFrameLock:SetChecked(Questo_Data.config.lockFrame);
	end);
	if Questo_Data.config.lockFrame == nil then
		Questo_Data.config.lockFrame = false;
	end
	QuestoOptionsPanel.LockFrame:SetChecked(Questo_Data.config.lockFrame);
	Questo_NPCFrameLock:SetChecked(Questo_Data.config.lockFrame);
	
	local textSpeedFactor = Questo_Data.config.textSpeedFactor or 0.5;
	local textSpeedTextSampleAnimation = 0;
	QuestoOptionsPanel.TextSpeed.Title:SetText(loc("Q_CONFIG_TEXTSPEED_TITLE"));
	QuestoOptionsPanel.TextSpeed.TextSample:SetText(loc("Q_CONFIG_BIG_SAMPLE_TEXT"));
	QuestoOptionsPanelTextSpeedSliderLow:SetText(loc("Q_CONFIG_TEXTSPEED_INSTANT"));
	QuestoOptionsPanelTextSpeedSliderHigh:SetText(loc("Q_CONFIG_TEXTSPEED_HIGH"));
	QuestoOptionsPanel.TextSpeed.Slider:SetScript("OnValueChanged", function(self, speed)
		QuestoOptionsPanelTextSpeedSliderText:SetText(loc("Q_CONFIG_TEXTSPEED"):format(speed));
		textSpeedFactor = speed;
		Questo_Data.config.textSpeedFactor = textSpeedFactor;
		textSpeedTextSampleAnimation = 0;
	end);
	QuestoOptionsPanel.TextSpeed.Slider:SetValue(textSpeedFactor);

	local ANIMATION_TEXT_SPEED = 80;
	local function onUpdateTextSpeedSample(self, elapsed)
		if textSpeedTextSampleAnimation == nil then return end
		textSpeedTextSampleAnimation = textSpeedTextSampleAnimation + (elapsed * (ANIMATION_TEXT_SPEED * Questo_Data.config.textSpeedFactor));
		if textSpeedFactor == 0 or textSpeedTextSampleAnimation >= QuestoOptionsPanel.TextSpeed.TextSample:GetText():len() then
			textSpeedTextSampleAnimation = nil;
			QuestoOptionsPanel.TextSpeed.TextSample:SetAlphaGradient(QuestoOptionsPanel.TextSpeed.TextSample:GetText():len(), 1);
		else
			QuestoOptionsPanel.TextSpeed.TextSample:SetAlphaGradient(textSpeedTextSampleAnimation, 30);
		end
	end

	QuestoOptionsPanel:SetScript("OnUpdate", onUpdateTextSpeedSample);



	
end
